//
//  ObviousEngine.h
//  ObviousEngine
//
//  Copyright 2012 Obvious Engine. All rights reserved.
//

#import <Foundation/Foundation.h>

typedef struct
{
    float x;
    float y;
    float z;
} OEVec3;

typedef struct
{
    float x;
    float y;
    float z;
    float w;
} OEQuaternion;

typedef struct
{
    float w;
    float h;
    BOOL isCylindrical;
    float cylinderRadius;
} OEModelInfo;

@interface ObviousEngine : NSObject
{
}

// Call this first to initialise Obvious Engine, passing in the
// licence data contained in your licence file.
- (id)initWithLicenceData:(const char* )data;

// Registers a model file, allowing it to be detected and tracked. 
// Up to 5 models can be registered with the system at once.
// 
// Returns an id which can be used for future operations.
// Returns -1 if the model could not be loaded.
- (int)addModel:(NSString *)path;

// De-registers a model with the engine.
// Pass the id returned from addModel.
- (BOOL)removeModel:(int)index;

// Find out information about a model.
// Pass the id returned from addModel.
- (OEModelInfo)modelInfo:(int)index;

// Set new frame data (grayscale).
// It is assumed that the image is 480x360 pixels, 1 byte per pixel.
// This data will be copied internally by the engine.
- (void)setFrameGrayscale:(unsigned char *)data;

// Set new frame data (BGRA).
// It is assumed that the image is 480x360 pixels, 4 bytes per pixel.
// This data will be copied internally by the engine.
- (void)setFrameBGRA:(unsigned char *)data;

// Performs all the tracking and detection work for the current frame.
// The frame data should be set first using setFrameGrayscale.
- (void)update;

// Query whether a model has been detected in the current frame.
- (BOOL)hasDetection;

// Returns the id of the currently active model.
// Only meaningful when hasDetection is true.
- (int)activeModel;

// Returns an OpenGL projection matrix with the correct field of view
// for the engine. If the viewport is set up to be 480x360, AR content
// rendered with this projection matrix will match the camera image.
- (const float *)projectionMatrix;

// Returns an OpenGL modelview matrix containing the rotation and translation
// of the model relative to the camera.
// Only meaningful when hasDetection is true.
- (const float *)modelViewMatrix;

// Returns the translation of the model relative to the camera.
// Only meaningful when hasDetection is true.
- (OEVec3)translation;

// Returns the rotation of the model relative to the camera.
// Only meaningful when hasDetection is true.
- (OEQuaternion)rotation;

// Returns the time taken in milliseconds for the last setFrame and update pair.
- (float)frameTime;

@end
