//
//  ARRenderer.mm
//  ObviousDemo
//
//  Copyright 2012 Obvious Engine. All rights reserved.
//

#include "ARRenderer.h"
#include "cylinder.h"
#include <cmath>

#include <glm/gtc/matrix_transform.hpp>

using namespace glm;
using namespace std;

ARRenderer::ARRenderer(int frameWidth, int frameHeight, const float* projectionMatrix, float stretchFactor) :
    m_frameWidth(frameWidth),
    m_frameHeight(frameHeight),
    m_halfFrameWidth(frameWidth>>1),
    m_halfFrameHeight(frameHeight>>1),
    m_stretchFactor(stretchFactor),
    m_activeBuffer(0),
    m_objectProgram(0),
    m_cameraProgram(0),
    m_warpProgram(0),
    m_uniformObjectModelview(-1),
    m_uniformObjectProjection(-1),
    m_uniformObjectColour(-1),
    m_uniformCameraProjection(-1),
    m_uniformCameraTextureY(-1),
    m_uniformCameraTextureUV(-1),
    m_hasDetection(false),
    m_modelview(1.f),
    m_scale(1.f),
    m_t(0),
    m_drawMode(0)
{
    mat4 P;
    memcpy(&P[0][0], projectionMatrix, 16*sizeof(float));
    
    // this matrix rotates the camera coordinate frame so it aligns with the
    // OpenGL frame. it also stretches one axis to account for the fact that
    // the camera image and screen have different aspect ratios
    m_cameraProjection = scale(mat4(1.f), vec3(stretchFactor, 1.f, 1.f)) * 
                         rotate(mat4(1.f), -90.f, vec3(0.f, 0.f, 1.f));
    m_objectProjection = m_cameraProjection * P;
    
    loadShaders();
    
    // allocate storage space for the two camera textures
    glGenTextures(4, m_frameTextures);
    // the first texture stores the Y component
    glActiveTexture(GL_TEXTURE0);
    for (int i = 0; i < 2; ++i)
    {
        glBindTexture(GL_TEXTURE_2D, m_frameTextures[i]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE, m_frameWidth, m_frameHeight, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, NULL);
    }
    // the second texture stores the UV components
    glActiveTexture(GL_TEXTURE1);
    for (int i = 0; i < 2; ++i)
    {
        glBindTexture(GL_TEXTURE_2D, m_frameTextures[2+i]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE_ALPHA, m_halfFrameWidth, m_halfFrameHeight, 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, NULL);
    }
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

ARRenderer::~ARRenderer()
{
    if (m_objectProgram)
    {
        glDeleteProgram(m_objectProgram);
        m_objectProgram = 0;
    }
    
    if (m_cameraProgram)
    {
        glDeleteProgram(m_cameraProgram);
        m_cameraProgram = 0;
    }
    glDeleteTextures(4, m_frameTextures);
}

void ARRenderer::setModelviewMatrix(const float* modelviewMatrix)
{
    memcpy(&m_modelview[0][0], modelviewMatrix, 16*sizeof(float));
}

void ARRenderer::setFrameImage(const unsigned char* y, const unsigned char* uv)
{
    // upload the camera image
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, m_frameTextures[1-m_activeBuffer]);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE, m_frameWidth, m_frameHeight, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, y);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, m_frameTextures[3-m_activeBuffer]);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE_ALPHA, m_halfFrameWidth, m_halfFrameHeight, 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, uv);
}

void ARRenderer::render()
{
    glClearColor(0.f, 0.f, 0.f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    
    // render the camera image as background
    renderCamera();
    
    if (m_hasDetection)
    {
        glEnable(GL_DEPTH_TEST);
        
        if (m_modelInfo.isCylindrical)
        {
            renderCylinder();
            if (m_drawMode == 0)
            {
                renderCylinderWarp();
            }
            else
            {
                renderCylinderParticles();
            }
            
        }
        else
        {
            if (m_drawMode == 0)
            {
                renderDisplacementGrid();
            }
            else
            {
                renderCube();
            }
        }
        
        glDisable(GL_DEPTH_TEST);
        ++m_t;
    }
    else
    {
        m_t = 0;
    }
}

void ARRenderer::renderCamera()
{
    static const GLfloat cameraVertices[] = {
        -1.f, -1.f,
         1.f, -1.f,
        -1.f,  1.f,
         1.f,  1.f,
    };
        
    glUseProgram(m_cameraProgram);
    
    glUniformMatrix4fv(m_uniformCameraProjection, 1, false, &m_cameraProjection[0][0]);
    
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, m_frameTextures[m_activeBuffer]);
    glUniform1i(m_uniformCameraTextureY, 0);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, m_frameTextures[2+m_activeBuffer]);
    glUniform1i(m_uniformCameraTextureUV, 1);
    
    glVertexAttribPointer(ATTRIBUTE_VERTEX, 2, GL_FLOAT, 0, 0, cameraVertices);
    glEnableVertexAttribArray(ATTRIBUTE_VERTEX);
    
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}

void ARRenderer::renderCube()
{
    static const GLfloat squareVertices[] = {
        -0.5f, -0.5f,
         0.5f, -0.5f,
        -0.5f,  0.5f,
         0.5f,  0.5f,
    };

    glUseProgram(m_objectProgram);
    glUniformMatrix4fv(m_uniformObjectProjection, 1, false, &m_objectProjection[0][0]);
    
    glVertexAttribPointer(ATTRIBUTE_VERTEX, 2, GL_FLOAT, 0, 0, squareVertices);
    glEnableVertexAttribArray(ATTRIBUTE_VERTEX);
    
    mat4 M = m_modelview;
    
    M = scale(M, vec3(m_scale*0.25f*m_modelInfo.w));
    M = translate(M, vec3(0.f, 0.f, 0.5f));
    mat4 Mi;
    // +x
    vec4 cxp(1.f, 0.f, 0.f, 1.0f);
    glUniform4fv(m_uniformObjectColour, 1, &cxp[0]);
    Mi = rotate(M, 90.f, vec3(0.f, 1.f, 0.f));
    Mi = translate(Mi, vec3(0.f, 0.f, 0.5f));
    glUniformMatrix4fv(m_uniformObjectModelview, 1, false, &Mi[0][0]);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    // -x
    vec4 cxn(0.f, 1.f, 1.f, 1.0f);
    glUniform4fv(m_uniformObjectColour, 1, &cxn[0]);
    Mi = rotate(M, -90.f, vec3(0.f, 1.f, 0.f));
    Mi = translate(Mi, vec3(0.f, 0.f, 0.5f));
    glUniformMatrix4fv(m_uniformObjectModelview, 1, false, &Mi[0][0]);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    // +y
    vec4 cyp(0.f, 1.f, 0.f, 1.0f);
    glUniform4fv(m_uniformObjectColour, 1, &cyp[0]);
    Mi = rotate(M, -90.f, vec3(1.f, 0.f, 0.f));
    Mi = translate(Mi, vec3(0.f, 0.f, 0.5f));
    glUniformMatrix4fv(m_uniformObjectModelview, 1, false, &Mi[0][0]);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    // -y
    vec4 cyn(1.f, 0.f, 1.f, 1.0f);
    glUniform4fv(m_uniformObjectColour, 1, &cyn[0]);
    Mi = rotate(M, 90.f, vec3(1.f, 0.f, 0.f));
    Mi = translate(Mi, vec3(0.f, 0.f, 0.5f));
    glUniformMatrix4fv(m_uniformObjectModelview, 1, false, &Mi[0][0]);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    // +z
    vec4 czp(0.f, 0.f, 1.f, 1.0f);
    glUniform4fv(m_uniformObjectColour, 1, &czp[0]);
    Mi = translate(M, vec3(0.f, 0.f, 0.5f));
    glUniformMatrix4fv(m_uniformObjectModelview, 1, false, &Mi[0][0]);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}

void ARRenderer::renderDisplacementGrid()
{
    static const GLfloat squareVertices[] = {
        -0.5f, -0.5f, 0.f,
         0.5f, -0.5f, 0.f,
        -0.5f,  0.5f, 0.f,
         0.5f,  0.5f, 0.f,
    };
    
    static const GLfloat squareNormals[] = {
        0.f, 0.f, 1.f,
        0.f, 0.f, 1.f,
        0.f, 0.f, 1.f,
        0.f, 0.f, 1.f,
    };
    
    mat4 M = m_modelview;
    M = scale(M, vec3(0.99f*m_modelInfo.w, 0.99f*m_modelInfo.h, m_scale*200.f));
    
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_STENCIL_TEST);
    
    glUseProgram(m_objectProgram);
    glUniformMatrix4fv(m_uniformObjectProjection, 1, false, &m_objectProjection[0][0]);
    vec4 cclear(0.f, 0.f, 0.f, 0.0f);
    glUniform4fv(m_uniformObjectColour, 1, &cclear[0]);
    glVertexAttribPointer(ATTRIBUTE_VERTEX, 3, GL_FLOAT, 0, 0, squareVertices);
    glEnableVertexAttribArray(ATTRIBUTE_VERTEX);
    
    glStencilFunc(GL_ALWAYS, 1, 1);
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
    glUniformMatrix4fv(m_uniformObjectModelview, 1, false, &M[0][0]);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    
    glEnable(GL_DEPTH_TEST);
    
    glStencilFunc(GL_NOTEQUAL, 1, 1);
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
    mat4 Mi = scale(M, vec3(50.f));
    glUniformMatrix4fv(m_uniformObjectModelview, 1, false, &Mi[0][0]);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    
    glDisable(GL_STENCIL_TEST);
    
    vec4 cblack(0.f, 0.f, 0.f, 1.0f);
    glUniform4fv(m_uniformObjectColour, 1, &cblack[0]);
    Mi = translate(M, vec3(0.f, 0.f, -2.f));
    Mi = scale(Mi, vec3(40.f));
    glUniformMatrix4fv(m_uniformObjectModelview, 1, false, &Mi[0][0]);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    
    glUseProgram(m_warpProgram);
    
    float tf = (float)m_t/15;
    glUniform1f(m_uniformWarpTime, tf);
    glUniform1i(m_uniformWarpIsCylindrical, 0);
    glUniform1f(m_uniformWarpStretchFactor, m_stretchFactor);
    
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, m_frameTextures[m_activeBuffer]);
    glUniform1i(m_uniformWarpTextureY, 0);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, m_frameTextures[2+m_activeBuffer]);
    glUniform1i(m_uniformWarpTextureUV, 1);
    
    glUniformMatrix4fv(m_uniformWarpProjection, 1, false, &m_objectProjection[0][0]);

    glVertexAttribPointer(ATTRIBUTE_DISPLACEMENT, 3, GL_FLOAT, 0, 0, squareNormals);
    glEnableVertexAttribArray(ATTRIBUTE_DISPLACEMENT);
    
    glVertexAttribPointer(ATTRIBUTE_VERTEX, 3, GL_FLOAT, 0, 0, squareVertices);
    glEnableVertexAttribArray(ATTRIBUTE_VERTEX);
    
    const int kNumX = 8;
    const int kNumY = 8;
    static float scales[kNumX*kNumY];
    static bool scalesSet = false;
    for (int x = 0; x < kNumX; ++x)
    {
        for (int y = 0; y < kNumY; ++y)
        {
            if (!scalesSet) scales[kNumX*y+x] = 1.f*(0.25f + 0.75f*(float)rand()/RAND_MAX);
            glUniform1f(m_uniformWarpScale, scales[kNumX*y+x]);
            
            Mi = scale(M, vec3(1.f/kNumX, 1.f/kNumY, 1.f));
            Mi = translate(Mi, vec3((float)x-(float)(kNumX-1)/2, (float)y-(float)(kNumY-1)/2, 0.f));
            
            glUniformMatrix4fv(m_uniformWarpModelview, 1, false, &Mi[0][0]);
            glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
        }
    }
    scalesSet = true;

    
}

// hacky visual tweaking, you will probably need to change this
static const float kRadiusTweak = 0.92f; 

void ARRenderer::renderCylinder()
{
    float radius = kRadiusTweak*m_modelInfo.cylinderRadius;
    
    // set up an invisible cylinder for depth masking
    mat4 M = translate(m_modelview, vec3(0.f, 0.f, -radius));
    M = rotate(M, 90.f, vec3(1.f, 0.f, 0.f));
    M = scale(M, vec3(radius, radius, 1.5f*m_modelInfo.h));

    glUseProgram(m_objectProgram);
    glUniformMatrix4fv(m_uniformObjectProjection, 1, false, &m_objectProjection[0][0]);
    glUniformMatrix4fv(m_uniformObjectModelview, 1, false, &M[0][0]);
    vec4 c(0.f, 0.f, 0.f, 0.0f);
    glUniform4fv(m_uniformObjectColour, 1, &c[0]);
    glVertexAttribPointer(ATTRIBUTE_VERTEX, 3, GL_FLOAT, 0, 0, cylinderVerts);
    glEnableVertexAttribArray(ATTRIBUTE_VERTEX);
    glDrawArrays(GL_TRIANGLES, 0, cylinderNumVerts);
}

void ARRenderer::renderCylinderWarp()
{
    float radius = kRadiusTweak*m_modelInfo.cylinderRadius;

    // set up the warping cylinder
    mat4 M = translate(m_modelview, vec3(0.f, 0.f, -radius));
    M = rotate(M, 90.f, vec3(1.f, 0.f, 0.f));
    M = scale(M, vec3(1.01f*radius, 1.01f*radius, m_modelInfo.h));
    
    float tf = ((float)m_t)/10;
    float s = 30.f;
    if (m_t < 30) s *= (float)m_t/30;
    
    glUseProgram(m_warpProgram);
    
    glUniform1f(m_uniformWarpTime, tf);
    glUniform1f(m_uniformWarpScale, m_scale*s);
    glUniform1i(m_uniformWarpIsCylindrical, 1);
    glUniform1f(m_uniformWarpStretchFactor, m_stretchFactor);
    
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, m_frameTextures[m_activeBuffer]);
    glUniform1i(m_uniformWarpTextureY, 0);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, m_frameTextures[2+m_activeBuffer]);
    glUniform1i(m_uniformWarpTextureUV, 1);
    
    glUniformMatrix4fv(m_uniformWarpProjection, 1, false, &m_objectProjection[0][0]);
    glUniformMatrix4fv(m_uniformWarpModelview, 1, false, &M[0][0]);

    glVertexAttribPointer(ATTRIBUTE_DISPLACEMENT, 3, GL_FLOAT, 0, 0, cylinderNormals);
    glEnableVertexAttribArray(ATTRIBUTE_DISPLACEMENT);
    
    glVertexAttribPointer(ATTRIBUTE_VERTEX, 3, GL_FLOAT, 0, 0, cylinderVerts);
    glEnableVertexAttribArray(ATTRIBUTE_VERTEX);
    
    glDrawArrays(GL_TRIANGLES, 0, cylinderNumVerts);
}

void ARRenderer::renderCylinderParticles()
{
    float radius = kRadiusTweak*m_modelInfo.cylinderRadius;
    
    static const GLfloat squareVertices[] = {
        -10.f, -10.f,
         10.f, -10.f,
        -10.f,  10.f,
         10.f,  10.f,
    };

    glUseProgram(m_objectProgram);
    glUniformMatrix4fv(m_uniformObjectProjection, 1, false, &m_objectProjection[0][0]);
    
    vec4 color(1.f, 1.f, 1.f, 1.0f);
    glUniform4fv(m_uniformObjectColour, 1, &color[0]);
    
    glVertexAttribPointer(ATTRIBUTE_VERTEX, 2, GL_FLOAT, 0, 0, squareVertices);
    glEnableVertexAttribArray(ATTRIBUTE_VERTEX);

    mat4 MM = translate(m_modelview, vec3(0.f, 0.f, -radius));
    int r = 0;
    MM = rotate(MM, (float)r, vec3(0.f, 1.f, 0.f));
    float tf = (float)m_t/15;

    const int n = 64*2;
    for (int i = 0; i < n; ++i)
    {
        mat4 M = rotate(MM, i*360.f/n, vec3(0.f, 1.f, 0.f));
        M = translate(M, vec3(1.0+abs(cos(tf+i))*radius+30.f, 0.f, 0.f));
        //glUniformMatrix4fv(m_uniformObjectModelview, 1, false, &Mi[0][0]);
        //glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
        mat4 Mi;
        // +x
        vec4 cxp(1.f, 1.f, 1.f, abs(std::cos(i+tf)*0.2));
        glUniform4fv(m_uniformObjectColour, 1, &cxp[0]);
        Mi = rotate(M, 90.f, vec3(0.f, 1.f, 0.f));
        glUniformMatrix4fv(m_uniformObjectModelview, 1, false, &Mi[0][0]);
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    }
}

////////////////////////////////////////////////////////////////////////////////
// Shader stuff
////////////////////////////////////////////////////////////////////////////////


bool ARRenderer::compileShader(GLuint* shader, GLenum type, const char* path)
{
    GLint status;
    const GLchar *source;
    
    NSString *file = [NSString stringWithCString:path encoding:NSUTF8StringEncoding];
    source = (GLchar *)[[NSString stringWithContentsOfFile:file encoding:NSUTF8StringEncoding error:nil] UTF8String];
    if (!source)
    {
        NSLog(@"Failed to load vertex shader");
        return false;
    }
    
    *shader = glCreateShader(type);
    glShaderSource(*shader, 1, &source, NULL);
    glCompileShader(*shader);
    
#if defined(DEBUG)
    GLint logLength;
    glGetShaderiv(*shader, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetShaderInfoLog(*shader, logLength, &logLength, log);
        NSLog(@"Shader compile log:\n%s", log);
        free(log);
    }
#endif
    
    glGetShaderiv(*shader, GL_COMPILE_STATUS, &status);
    if (status == 0)
    {
        glDeleteShader(*shader);
        return false;
    }
    
    return true;
}

bool ARRenderer::linkProgram(GLuint prog)
{
    GLint status;
    
    glLinkProgram(prog);
    
#if defined(DEBUG)
    GLint logLength;
    glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetProgramInfoLog(prog, logLength, &logLength, log);
        NSLog(@"Program link log:\n%s", log);
        free(log);
    }
#endif
    
    glGetProgramiv(prog, GL_LINK_STATUS, &status);
    if (status == 0)
        return false;
    
    return true;
}

bool ARRenderer::validateProgam(GLuint prog)
{
    GLint logLength, status;
    
    glValidateProgram(prog);
    glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetProgramInfoLog(prog, logLength, &logLength, log);
        NSLog(@"Program validate log:\n%s", log);
        free(log);
    }
    
    glGetProgramiv(prog, GL_VALIDATE_STATUS, &status);
    if (status == 0)
        return false;
    
    return true;
}

bool ARRenderer::loadShaders()
{
    // Create object shader program.
    m_objectProgram = glCreateProgram();
    
    if (!loadShaderPair(m_objectProgram, 
        [[[NSBundle mainBundle] pathForResource:@"ObjectShader" ofType:@"vsh"] UTF8String], 
        [[[NSBundle mainBundle] pathForResource:@"ObjectShader" ofType:@"fsh"] UTF8String]))
    {
        NSLog(@"Failed to compile object shader program");
        if (m_objectProgram)
        {
            glDeleteProgram(m_objectProgram);
            m_objectProgram = 0;
        }
        return false;
    }
    
    // Get uniform locations.
    m_uniformObjectModelview = glGetUniformLocation(m_objectProgram, "modelviewMatrix");
    m_uniformObjectProjection = glGetUniformLocation(m_objectProgram, "projectionMatrix");
    m_uniformObjectColour = glGetUniformLocation(m_objectProgram, "colour");
    
    // Create camera shader program.
    m_cameraProgram = glCreateProgram();
    
    if (!loadShaderPair(m_cameraProgram, 
        [[[NSBundle mainBundle] pathForResource:@"CameraShader" ofType:@"vsh"] UTF8String], 
        [[[NSBundle mainBundle] pathForResource:@"CameraShader" ofType:@"fsh"] UTF8String]))
    {
        NSLog(@"Failed to compile camera shader program");
        if (m_cameraProgram)
        {
            glDeleteProgram(m_cameraProgram);
            m_cameraProgram = 0;
        }
        return false;
    }
    
    // Get uniform locations.
    m_uniformCameraTextureY = glGetUniformLocation(m_cameraProgram, "textureY");  
    m_uniformCameraTextureUV = glGetUniformLocation(m_cameraProgram, "textureUV");  
    m_uniformCameraProjection = glGetUniformLocation(m_cameraProgram, "projectionMatrix"); 
    
    // Create warp shader program
    m_warpProgram = glCreateProgram();
    
    if (!loadShaderPair(m_warpProgram, 
        [[[NSBundle mainBundle] pathForResource:@"WarpShader" ofType:@"vsh"] UTF8String], 
        [[[NSBundle mainBundle] pathForResource:@"WarpShader" ofType:@"fsh"] UTF8String]))
    {
        NSLog(@"Failed to compile warp shader program");
        if (m_warpProgram)
        {
            glDeleteProgram(m_warpProgram);
            m_warpProgram = 0;
        }
        return false;
    }
    
    // Get uniform locations.
    m_uniformWarpModelview = glGetUniformLocation(m_warpProgram, "modelviewMatrix");
    m_uniformWarpProjection = glGetUniformLocation(m_warpProgram, "projectionMatrix");
    m_uniformWarpStretchFactor = glGetUniformLocation(m_warpProgram, "stretchFactor");
    m_uniformWarpTextureY = glGetUniformLocation(m_warpProgram, "textureY");  
    m_uniformWarpTextureUV = glGetUniformLocation(m_warpProgram, "textureUV");
    m_uniformWarpTime = glGetUniformLocation(m_warpProgram, "time"); 
    m_uniformWarpScale = glGetUniformLocation(m_warpProgram, "scale"); 
    m_uniformWarpIsCylindrical = glGetUniformLocation(m_warpProgram, "isCylindrical");
    
    return true;
}

bool ARRenderer::loadShaderPair(GLuint program, const char* vertShaderPath, const char* fragShaderPath)
{
    GLuint vertShader, fragShader;
    
    // Create and compile vertex shader.
    if (!compileShader(&vertShader, GL_VERTEX_SHADER, vertShaderPath))
    {
        NSLog(@"Failed to compile vertex shader");
        return FALSE;
    }
    
    // Create and compile fragment shader.
    if (!compileShader(&fragShader, GL_FRAGMENT_SHADER, fragShaderPath))
    {
        NSLog(@"Failed to compile fragment shader");
        return FALSE;
    }
    
    // Attach vertex shader to program.
    glAttachShader(program, vertShader);
    
    // Attach fragment shader to program.
    glAttachShader(program, fragShader);
    
    // Bind attribute locations.
    glBindAttribLocation(program, ATTRIBUTE_VERTEX, "position");
    glBindAttribLocation(program, ATTRIBUTE_DISPLACEMENT, "displacement");
        
    // Link program.
    if (!linkProgram(program))
    {
        NSLog(@"Failed to link program: %d", program);
        
        if (vertShader)
        {
            glDeleteShader(vertShader);
            vertShader = 0;
        }
        if (fragShader)
        {
            glDeleteShader(fragShader);
            fragShader = 0;
        }
        return false;
    }
    
    // Release vertex and fragment shaders.
    if (vertShader)
        glDeleteShader(vertShader);
    if (fragShader)
        glDeleteShader(fragShader); 
    
    return true;
}

