//
//  ViewController.m
//  ObviousDemo
//
//  Copyright 2012 Obvious Engine. All rights reserved.
//

#import "ViewController.h"

#include "ARRenderer.h"

#import "ObviousEngineLicence.h"

// this is the resolution of the camera image when using 
// AVCaptureSessionPresetMedium, anything else is untested 
// so don't change this!
static const int kFrameWidth = 480;
static const int kFrameHeight = 360;

#if !USE_CAMERA
static bool imageLoaded = false;
static unsigned char imageDataY[kFrameWidth*kFrameHeight*sizeof(char)];
static unsigned char imageDataUV[kFrameWidth*(kFrameHeight/2)*sizeof(char)];
#endif

@interface ViewController () 
@property (strong, nonatomic) EAGLContext *context;
@property (strong, nonatomic) GLKBaseEffect *effect;
@property (nonatomic, retain) NSArray *modelFiles;
- (void)loadNextModel;
- (void)hideModelLabel;
- (void)toggleDrawMode;
- (void)handlePinchGesture:(UIPinchGestureRecognizer *)gestureRecognizer;
#if USE_CAMERA
- (void)captureOutput:(AVCaptureOutput *)captureOutput didOutputSampleBuffer:(CMSampleBufferRef)sampleBuffer fromConnection:(AVCaptureConnection *)connection;
#endif
@end

@implementation ViewController

@synthesize context = _context;
@synthesize effect = _effect;
@synthesize modelLabel;
@synthesize modelFiles;

- (void)dealloc
{
    [_context release];
    [_effect release];
    [super dealloc];
}

- (void)viewDidLoad
{
    [super viewDidLoad];
    
    self.context = [[[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES2] autorelease];
    if (!self.context) {
        NSLog(@"Failed to create ES context");
    }
    self.preferredFramesPerSecond = 30;
    
    GLKView *view = (GLKView *)self.view;
    view.context = self.context;
    view.drawableDepthFormat = GLKViewDrawableDepthFormat24;
    view.drawableStencilFormat = GLKViewDrawableStencilFormat8;
    view.drawableDepthFormat = GLKViewDrawableDepthFormat16;
    // TODO: because the YUV->RGB conversion is currently being done in a pixel 
    // shader, enabling retina graphics causes a big performance hit. 
    // a solution to this would be to use a framebuffer and render to 
    // texture to do the conversion, which should be much more efficient.
    view.contentScaleFactor = 1.f;
    
    [EAGLContext setCurrentContext:self.context];
    
    arSystemLock = [[NSLock alloc] init];
    arSystem = [[ObviousEngine alloc] initWithLicenceData:kOELicenceData];
    
    // stretch factor accounts for the fact that the camera and the screen
    // have different aspect ratios
    float screenAspectRatio = (float)self.view.frame.size.height/self.view.frame.size.width;
    float cameraAspectRatio = (float)kFrameWidth/kFrameHeight;
    float stretchFactor = screenAspectRatio/cameraAspectRatio;
    renderer = new ARRenderer(kFrameWidth, kFrameHeight, [arSystem projectionMatrix], stretchFactor);
    
    // find any model files which have been added via iTunes file sharing
    NSString *documentDir = [NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES) objectAtIndex:0];
    NSArray *dirContents = [[NSFileManager defaultManager] contentsOfDirectoryAtPath:documentDir error:nil];
    self.modelFiles = [dirContents filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"self ENDSWITH '.model'"]];
    activeModel = -1;
    if (modelFiles.count == 0)
    {
        UIAlertView *alert = [[UIAlertView alloc] initWithTitle:@"Error" message:@"No model files found, please install using iTunes file sharing." delegate:nil cancelButtonTitle:@"OK" otherButtonTitles: nil];
        [alert show];
        [alert release];
    }
    else
    {
        // load the first model. subsequent models can be loaded by swiping.
        [self loadNextModel];
        UISwipeGestureRecognizer *swipeLR = [[UISwipeGestureRecognizer alloc] initWithTarget:self action:@selector(loadNextModel)];
        UISwipeGestureRecognizer *swipeUD = [[UISwipeGestureRecognizer alloc] initWithTarget:self action:@selector(loadNextModel)];
        swipeLR.direction = (UISwipeGestureRecognizerDirection)(UISwipeGestureRecognizerDirectionRight | UISwipeGestureRecognizerDirectionLeft);
        swipeUD.direction = (UISwipeGestureRecognizerDirection)(UISwipeGestureRecognizerDirectionUp | UISwipeGestureRecognizerDirectionDown);
        [self.view addGestureRecognizer:swipeLR];
        [self.view addGestureRecognizer:swipeUD];
        [swipeLR release];
        [swipeUD release];
    }
    
    UIPinchGestureRecognizer *pinch = [[UIPinchGestureRecognizer alloc] initWithTarget:self action:@selector(handlePinchGesture:)];
    [self.view addGestureRecognizer:pinch];
    [pinch release];
    
    UITapGestureRecognizer *tap = [[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(toggleDrawMode)];
    tap.numberOfTapsRequired = 2;
    [self.view addGestureRecognizer:tap];
    [tap release];
    
#if USE_CAMERA
    // set up camera
    captureSession = [[AVCaptureSession alloc] init];
    captureQueue = dispatch_queue_create("ar_capture", NULL);
    cameraDevice = [AVCaptureDevice defaultDeviceWithMediaType:AVMediaTypeVideo];
    AVCaptureDeviceInput *videoInput = [AVCaptureDeviceInput deviceInputWithDevice:cameraDevice error:NULL];
    dataOutput = [[AVCaptureVideoDataOutput alloc] init];
    if (videoInput && dataOutput)
    {
        // note that the camera is set up to deliver data in YUV format.
        // this is good for the AR system, since the first plane is a
        // grayscale image, and we don't have to convert RGB->grayscale.
        // however, it requires a shader to reassemble the planes
        // into an RGB image for rendering.
        NSDictionary* videoSettings = [NSDictionary 
                                       dictionaryWithObject:
                                       [NSNumber numberWithUnsignedInt:kCVPixelFormatType_420YpCbCr8BiPlanarVideoRange]
                                       forKey:(NSString*)kCVPixelBufferPixelFormatTypeKey]; 
        dataOutput.videoSettings = videoSettings;
        
        [dataOutput setSampleBufferDelegate:self queue:captureQueue];
        
        dataOutput.alwaysDiscardsLateVideoFrames = YES;
        [captureSession beginConfiguration]; 
        [captureSession addInput:videoInput];
        [captureSession addOutput:dataOutput];
        
        // set the capture framerate
        AVCaptureConnection *conn = [dataOutput connectionWithMediaType:AVMediaTypeVideo];
        if (conn.supportsVideoMinFrameDuration)
            conn.videoMinFrameDuration = CMTimeMake(1, 30);
        if (conn.supportsVideoMaxFrameDuration)
            conn.videoMaxFrameDuration = CMTimeMake(1, 30);
        
        // this preset delivers 480x360 frames, which is what Obvious Engine
        // is tuned to use, so don't change this...
        [captureSession setSessionPreset:AVCaptureSessionPresetMedium];
        [captureSession commitConfiguration];
        [captureSession startRunning];
    }
#endif
}

- (void)viewDidUnload
{    
    [super viewDidUnload];
    
#if USE_CAMERA
    [captureSession stopRunning];
    [dataOutput release];
    [captureSession release];
    dispatch_release(captureQueue);
#endif

    [arSystem release];
    [arSystemLock release];
    delete renderer;
    renderer = 0;
    
    self.effect = nil;
    
    if ([EAGLContext currentContext] == self.context) {
        [EAGLContext setCurrentContext:nil];
    }
    self.context = nil;
}

- (void)didReceiveMemoryWarning
{
    [super didReceiveMemoryWarning];
    // Release any cached data, images, etc. that aren't in use.
}


- (void)loadNextModel
{
    if (modelFiles.count == 0) return;
    
    [arSystemLock lock];
    
    renderer->setHasDetection(false);
    
    if (activeModel >= 0)
    {
        // unload existing model
        [arSystem removeModel:modelHandle];
    }
    
    activeModel = (activeModel + 1) % modelFiles.count;
    NSString *modelFile = [modelFiles objectAtIndex:activeModel];
    NSString *documentDir = [NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES) objectAtIndex:0];
    NSString *path = [documentDir stringByAppendingPathComponent:modelFile];
    modelLabel.text = modelFile;
    modelLabel.hidden = NO;
    modelLabel.alpha = 1.0;
    
    if (modelLabelTimer)
    {
        [modelLabelTimer invalidate];
        [modelLabelTimer release];
    }
    modelLabelTimer = [[NSTimer scheduledTimerWithTimeInterval:2.0 target:self selector:@selector(hideModelLabel) userInfo:nil repeats:NO] retain];
    
    modelHandle = [arSystem addModel:path];
    if (modelHandle == -1)
    {
        NSString *msg = [NSString stringWithFormat:@"Could not load model file: %@", modelFile];
        UIAlertView *alert = [[UIAlertView alloc] initWithTitle:@"Error" message:msg delegate:nil cancelButtonTitle:@"OK" otherButtonTitles: nil];
        [alert show];
        [alert release];
    }
    else
    {
        renderer->setModelInfo([arSystem modelInfo:modelHandle]);
    }
         
    
    [arSystemLock unlock];
}

- (void)toggleDrawMode
{
    renderer->toggleDrawMode();
}

- (void)hideModelLabel
{
    [UIView animateWithDuration:0.3
     animations:^{modelLabel.alpha = 0.0;}
     completion:^(BOOL finished){ modelLabel.hidden = YES; }];
}

- (void)handlePinchGesture:(UIPinchGestureRecognizer *)gestureRecognizer
{
    static float startScale;
    if (gestureRecognizer.state == UIGestureRecognizerStateBegan)
    {
        startScale = renderer->getScale();
    }
    float scale = gestureRecognizer.scale*startScale;
    scale = (scale > 10.f) ? 10.f : (scale < 0.1f) ? 0.1f : scale;
    renderer->setScale(scale);
}


#pragma mark - GLKView and GLKViewController delegate methods

- (void)update
{
#if !USE_CAMERA
    if (!imageLoaded)
    {
        // load a static image to run in the simulator
        NSData *data = [NSData dataWithContentsOfFile:[[NSBundle mainBundle] pathForResource:@"frame" ofType:@"bin"]];
        memcpy(imageDataY, [data bytes] , kFrameWidth*kFrameHeight);
        memcpy(imageDataUV, (unsigned char*)[data bytes]+kFrameWidth*kFrameHeight, kFrameWidth*kFrameHeight/2);
        
        [EAGLContext setCurrentContext:self.context];
        renderer->setFrameImage(imageDataY, imageDataUV);
        renderer->flipTextureBuffers();
        imageLoaded = true;
    }
    [arSystem setFrameGrayscale:imageDataY];
    [arSystem update];
    renderer->setHasDetection([arSystem hasDetection]);
    renderer->setModelviewMatrix([arSystem modelViewMatrix]);
#endif
}

- (void)glkView:(GLKView *)view drawInRect:(CGRect)rect
{
    [EAGLContext setCurrentContext:self.context];
    renderer->render();
}

#if USE_CAMERA
// this is the camera callback
- (void)captureOutput:(AVCaptureOutput *)captureOutput didOutputSampleBuffer:(CMSampleBufferRef)sampleBuffer fromConnection:(AVCaptureConnection *)connection
{
    // get a handle on the camera image data
    CVImageBufferRef imageBuffer = CMSampleBufferGetImageBuffer(sampleBuffer); 
    CVPixelBufferLockBaseAddress(imageBuffer, 0); 
    uint8_t *baseAddressY = (uint8_t *)CVPixelBufferGetBaseAddressOfPlane(imageBuffer, 0);
    uint8_t *baseAddressUV = (uint8_t *)CVPixelBufferGetBaseAddressOfPlane(imageBuffer, 1);

// uncomment the following to dump a camera frame for loading in the simulator
// you can then get the frame from the device, and add it as a resource to this
// app for loading in the simulator
//    NSString *path = [NSTemporaryDirectory() stringByAppendingPathComponent:@"frame.bin"];
//    FILE *f = fopen([path UTF8String], "wb");
//    fwrite(baseAddressY, 1, kFrameWidth*kFrameHeight, f);
//    fwrite(baseAddressUV, 1, kFrameWidth*kFrameHeight/2, f);
//    fclose(f);
        
    // pass only the grayscale component of the frame to the AR system
    [arSystem setFrameGrayscale:baseAddressY];
    
    // upload the camera image to the renderer.
    // note we do this from the main thread.
    dispatch_sync(dispatch_get_main_queue(), ^{
        [EAGLContext setCurrentContext:self.context];
        renderer->setFrameImage(baseAddressY, baseAddressUV);
    });
    
    CVPixelBufferUnlockBaseAddress(imageBuffer,0);
    
    [arSystemLock lock];
    
    // perform the processing. note this happens after we've unlocked the
    // image buffer, freeing it up for the camera.
    [arSystem update];

    //NSLog(@"frame time: %f", [arSystem frameTime]);
    
    if ([arSystem hasDetection])
    {
        renderer->setModelviewMatrix([arSystem modelViewMatrix]);
        renderer->setHasDetection(true);
        lostFrames = 0;
        ++trackedFrames;
        if ((trackedFrames == 50) && (cameraDevice.exposureMode != AVCaptureExposureModeLocked))
        {
            // lock camera settings. this is useful as otherwise the autofocus
            // kicks in and can kill tracking.
            //NSLog(@"locking camera settings");
            [cameraDevice lockForConfiguration:NULL];
            cameraDevice.exposureMode = AVCaptureExposureModeLocked;
            // this is only supported on cameras with focus
            if ([cameraDevice isFocusModeSupported:AVCaptureFocusModeContinuousAutoFocus])
                cameraDevice.focusMode = AVCaptureFocusModeLocked;
            [cameraDevice unlockForConfiguration];
        }
    }
    else if (++lostFrames > 2)
    {
        renderer->setHasDetection(false);
        trackedFrames = 0;
        objectScale = renderer->getScale();
        if ((lostFrames == 25) && (cameraDevice.exposureMode == AVCaptureExposureModeLocked))
        {
            // unlock camera settings
            //NSLog(@"unlocking camera settings");
            [cameraDevice lockForConfiguration:NULL];
            cameraDevice.exposureMode = AVCaptureExposureModeContinuousAutoExposure;
            if ([cameraDevice isFocusModeSupported:AVCaptureFocusModeContinuousAutoFocus])
                cameraDevice.focusMode = AVCaptureFocusModeContinuousAutoFocus;
        }
    }
    renderer->flipTextureBuffers();
    [arSystemLock unlock];
}
#endif

@end
